#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>

#include <string.h>
#include <time.h>
#include <sys/timeb.h>

#include "ScanDBMine.h"
#include "FPtree.h"
#include "parameters.h"
#include "fsout.h"
#include "data.h"
#include "inline_routine.h"

FPNODE_BUF gofpnode_buf;
HEADER_TABLE gpheader_table;

HEADER_TABLE gpdfs_header_array;
int gndfs_header_size;
int gndfs_header_pos;

TAIL_NODE *gpfree_tail_nodes;
TAIL_NODE *gptailnodes_head;
TAIL_NODE *gptailnodes_tail;
PAT_SET gplocal_fullitems;
int *gplocal_item_sup_map;
int *gplocal_item_sumsup_map;
CParameters goparameters;


int main(int argc, char *argv[])
{
	int narg_no;

	if(argc!=3 && argc!=4)
	{
		printf("Usage\n");
		printf("\t%s  data_filename  data_type nmin_sup(an obsolute number) [output_filename]\n", argv[0]);
		return 0;
	}	

	narg_no = 1;
	strcpy(goparameters.szdata_filename, argv[narg_no]);
	narg_no++;

//	goparameters.ntype = atoi(argv[narg_no]);
//	narg_no++;

	goparameters.nmin_sup = atoi(argv[narg_no]);
	narg_no++;
	gnmin_sup = goparameters.nmin_sup;
	if(goparameters.nmin_sup<1)
	{
		printf("Please specify a native number for the minimum support threshold\n");
		return 0;
	}

	if(argc==narg_no+1)
	{
		goparameters.bresult_name_given = true;
		strcpy(goparameters.szresult_filename, argv[narg_no]);
	}
	else 
		goparameters.bresult_name_given = false;


	MineClosePats();

	_CrtDumpMemoryLeaks();

	return gntotal_closed;
}

void MineClosePats()
{
	FP_NODE *proot;
	HEADER_TABLE pheader_table;
	int *pitem_sup_map, ncapacity, i, nfreq_item, *pitem_order_map;

	struct timeb start, end;

	ftime(&start);

	if(goparameters.bresult_name_given)
		gpfout = new FSout(goparameters.szresult_filename);

	gpdata = new Data(goparameters.szdata_filename);

	gnmax_pattern_len = 0;
	gntotal_call = 0;
	gdused_mem_size = 0;
	gdmax_used_mem_size = 0;
	gntree_max_size = 0;
	gntotal_closed = 0;

	//count frequent items in original database
	pitem_sup_map = NULL;
	ncapacity = goDBMiner.ScanDBCountFreqItems(pitem_sup_map);

	gpprefix_itemset = NewIntArray(gnmax_trans_len);
	gnprefix_len = 0;

	//enumerate frequent itemsets
	gntotal_freqitems  = 0;
	for(i=0;i<gnmax_item_id;i++)
	{
		if(pitem_sup_map[i]==gndb_size)
		{
			gpprefix_itemset[gnprefix_len++] = i;
		}
		else if(pitem_sup_map[i]>=gnmin_sup)
		{
			gntotal_freqitems++;
			nfreq_item = i;
		}			
	}

	gplocal_item_sup_map = NewIntArray(gntotal_freqitems);
	gplocal_item_sumsup_map = NewIntArray(gntotal_freqitems);
	gplocal_fullitems.phead = NewPatPage();
	gptailnodes_head = NewTailNode();

	OutputOneClosedPat(gndb_size);

	if(gntotal_freqitems==0)
	{
		gnmax_pattern_len = 0;
		delete gpdata;
		DelIntArray(pitem_sup_map, ncapacity);
	}
	else if(gntotal_freqitems==1)
	{
		gpprefix_itemset[gnprefix_len++] = nfreq_item;
		OutputOneClosedPat(pitem_sup_map[nfreq_item]);
		gnprefix_len--;

		gnmax_pattern_len = 1;
		delete gpdata;
		DelIntArray(pitem_sup_map, ncapacity);
	}
	else if(gntotal_freqitems>1)
	{
		pheader_table = NewHeaderTable(gntotal_freqitems);
		gntotal_freqitems = 0;
		for(i=0;i<gnmax_item_id;i++)
		{
			if(pitem_sup_map[i]>=gnmin_sup && pitem_sup_map[i]<gndb_size)
			{
				pheader_table[gntotal_freqitems].nitem = i;
				pheader_table[gntotal_freqitems].nsupport = pitem_sup_map[i];
				pheader_table[gntotal_freqitems].pconddb = NULL;
				gntotal_freqitems++;
			}
		}
		qsort(pheader_table, gntotal_freqitems, sizeof(HEADER_NODE), comp_hnode_freq_des);
		DelIntArray(pitem_sup_map, ncapacity);

		gpheader_table = pheader_table;

		pitem_order_map = NewIntArray(gnmax_item_id, -1);
		for(i=0;i<gntotal_freqitems;i++)
			pitem_order_map[pheader_table[i].nitem] =i;

		gpdfs_item_counter = NewItemCounter(gntotal_freqitems);

		Init();

		gnum_of_newfreqitems = gntotal_freqitems;

		proot = goDBMiner.ScanDBBuildFPtree(pheader_table, pitem_order_map);
		gntree_init_size = sizeof(FPNODE_BUF)+gofpnode_buf.ntotal_pages*(sizeof(FP_NODE)*FPNODE_PAGE_SIZE+sizeof(FPNODE_PAGE));
		delete gpdata;

		gpdfs_item_suporder_map = NewIntArray(gntotal_freqitems);
		goFPtree.DepthFGGrowth(proot, pheader_table, gntotal_freqitems, 0);
		DelIntArray(gpdfs_item_suporder_map, gntotal_freqitems);

		gntree_max_size = sizeof(FPNODE_BUF)+gofpnode_buf.ntotal_pages*(sizeof(FP_NODE)*FPNODE_PAGE_SIZE+sizeof(FPNODE_PAGE));
		
		Destroy();
		DelIntArray(pitem_order_map, gnmax_item_id);
		DelItemCounter(gpdfs_item_counter, gntotal_freqitems);
		DelHeaderTable(pheader_table, gntotal_freqitems);
	}

	DelIntArray(gpprefix_itemset, gnmax_trans_len);

	DelIntArray(gplocal_item_sup_map, gntotal_freqitems);
	DelIntArray(gplocal_item_sumsup_map, gntotal_freqitems);
	DelPatSet(&gplocal_fullitems);
	DelTailNodes(gptailnodes_head);

	if(goparameters.bresult_name_given)
		delete gpfout;
	ftime(&end);
	gdtotal_running_time = end.time-start.time+(double)(end.millitm-start.millitm)/1000;

	PrintSummary();

}


void FPtree::DepthFGGrowth(FP_NODE *proot, HEADER_TABLE pheader_table, int num_of_freqitems, int num_of_tail_items)
{
	HEADER_TABLE pnewheader_table;
	FP_NODE *pnewroot, *pfpnode;
	FPNODE_PAGE *pstart_page;
	int k, i, nstart_pos, ncount;
	int num_of_tail_comm, num_of_comm_items, num_of_newfreqitems, num_of_newtail_items, num_of_newtail_comm;

	gntotal_call++;

	for(k=num_of_freqitems-1;k>=0; k--)
	{
		gpprefix_itemset[gnprefix_len] = pheader_table[k].nitem;
		gnprefix_len++;


		if(pheader_table[k].pconddb==NULL)
		{
			OutputOneClosedPat(pheader_table[k].nsupport);
		}
		else //if(pheader_table[k].pconddb!=NULL)
		{
			if(pheader_table[k].pconddb->pnode_link==NULL)
			{
				ncount = pheader_table[k].pconddb->frequency;
				if(ncount<pheader_table[k].nsupport)
					OutputOneClosedPat(pheader_table[k].nsupport);

				pfpnode = pheader_table[k].pconddb->pparent;
				if(ncount>=gnmin_sup)
				{
					num_of_tail_comm = CountTailFullItems(pheader_table[k].pconddb, num_of_freqitems+num_of_tail_items, k);
					if(num_of_tail_comm==0)
					{
						num_of_comm_items = 0;
						
						while(pfpnode!=NULL)
						{
							gpprefix_itemset[gnprefix_len++] = pheader_table[pfpnode->nitem_order].nitem;
							num_of_comm_items++;
							pfpnode = pfpnode->pparent;
						}
						if(num_of_comm_items>0 || ncount==pheader_table[k].nsupport)
							OutputOneClosedPat(ncount);
						gnprefix_len -= num_of_comm_items;
					}
				}
			}
			else
			{
				//count frequent items from FP-tree
				memset(gpdfs_item_suporder_map, 0, sizeof(int)*k);
				memset(&(gpdfs_item_suporder_map[k]), -1, sizeof(int)*(num_of_freqitems+num_of_tail_items-k));
				CountFreqItems(pheader_table, k, gpdfs_item_suporder_map);

				num_of_newfreqitems = 0;
				num_of_newtail_items = 0;
				num_of_comm_items = 0;
				for(i=0;i<k;i++)
				{
					if(gpdfs_item_suporder_map[i]==pheader_table[k].nsupport)
					{
						gpprefix_itemset[gnprefix_len++] = pheader_table[i].nitem;
						num_of_comm_items++;
					}
					else if(gpdfs_item_suporder_map[i]>=gnmin_sup)
					{
						gpdfs_item_counter[num_of_newfreqitems].nsupport = gpdfs_item_suporder_map[i];
						gpdfs_item_counter[num_of_newfreqitems].nitem = pheader_table[i].nitem;
						gpdfs_item_counter[num_of_newfreqitems].norder = i;
						num_of_newfreqitems++;
					}
					gpdfs_item_suporder_map[i] = -1;
				}
				if(num_of_newfreqitems>0)
					num_of_tail_comm = CountTailFullItems(pheader_table[k].pconddb, pheader_table[k].nsupport, num_of_freqitems+num_of_tail_items, k, true);
				else 
					num_of_tail_comm = CountTailFullItems(pheader_table[k].pconddb, pheader_table[k].nsupport, num_of_freqitems+num_of_tail_items, k, false);
				if(num_of_tail_comm==0)
				{
					OutputOneClosedPat(pheader_table[k].nsupport);

					if(num_of_newfreqitems>1)
					{
						qsort(gpdfs_item_counter, num_of_newfreqitems, sizeof(ITEM_COUNTER), comp_item_freq_des);
						for(i=0;i<num_of_newfreqitems;i++)
							gpdfs_item_suporder_map[gpdfs_item_counter[i].norder] = i;
					}

					for(i=k+1;i<num_of_freqitems+num_of_tail_items;i++)
					{
						if(gplocal_item_sumsup_map[i]>=gnmin_sup)
						{
							gpdfs_item_suporder_map[i] = num_of_newfreqitems+num_of_newtail_items;
							gpdfs_item_counter[num_of_newfreqitems+num_of_newtail_items].nitem = pheader_table[i].nitem;
							gpdfs_item_counter[num_of_newfreqitems+num_of_newtail_items].nsupport = 0;
							num_of_newtail_items++;
						}
					}

					if(num_of_newfreqitems==1)
					{
						gpprefix_itemset[gnprefix_len++] = gpdfs_item_counter[0].nitem;
						num_of_newtail_comm = GetTailFullItems(pheader_table[k].pconddb, gpdfs_item_counter[0].nsupport, num_of_freqitems+num_of_tail_items, gpdfs_item_counter[0].norder, gpdfs_item_suporder_map);
						if(num_of_newtail_comm==0)
							OutputOneClosedPat(gpdfs_item_counter[0].nsupport);
						gnprefix_len--;
					}
					else if(num_of_newfreqitems>1)
					{
						pnewheader_table = NewHeaderTable(num_of_newfreqitems+num_of_newtail_items);
						for(i=0;i<num_of_newfreqitems+num_of_newtail_items;i++)
						{
							pnewheader_table[i].nitem = gpdfs_item_counter[i].nitem;
							pnewheader_table[i].nsupport = gpdfs_item_counter[i].nsupport;
							pnewheader_table[i].pconddb = NULL;
						}

						gnum_of_newfreqitems = num_of_newfreqitems+num_of_newtail_items;
						pstart_page = gofpnode_buf.pcur_page;
						nstart_pos = gofpnode_buf.ncur_pos;
						pnewroot = BuildNewTreeWTail(pheader_table[k].pconddb, pnewheader_table, gpdfs_item_suporder_map); 
						//pnewroot = BuildNewFPTree(pheader_table[k].pconddb, pnewheader_table, gpdfs_item_suporder_map); 

						DepthFGGrowth(pnewroot, pnewheader_table, num_of_newfreqitems, num_of_newtail_items);

						Reset(pstart_page, nstart_pos);
						DelHeaderTable(pnewheader_table, num_of_newfreqitems+num_of_newtail_items);
					}
				}
				gnprefix_len -= num_of_comm_items;
			}
		}
		gnprefix_len--;
	}
}



void PrintSummary()
{
	printf("#closed itemsets: %d\n", gntotal_closed);
	if(gdused_mem_size!=0)
		printf("Error with memory: %.2f are not released\n", gdused_mem_size);

	FILE *fp_sum;

	fp_sum = fopen("fp-closed.sum.txt", "a+");
	if(fp_sum == NULL)
	{
		printf("Error[PrintSummary]: cannot open file fg.sum.txt\n");
		return;
	}
	fprintf(fp_sum, "fpc%s ", goparameters.szdata_filename);
	fprintf(fp_sum, "%f ", (double)goparameters.nmin_sup*100/gndb_size);
	fprintf(fp_sum, "%d %d\t", gnmax_pattern_len, gntotal_closed);
	fprintf(fp_sum, "%.2f\t", gdtotal_running_time);
	fprintf(fp_sum, "%.2fMB %.2fMB %.2fMB\t", gdmax_used_mem_size/(1<<20), (double)gntree_init_size/(1<<20), (double)gntree_max_size/(1<<20));

	fprintf(fp_sum, "\n");
	fclose(fp_sum);

}

